/*************************************************************************/
/*                                                                       */
/*                Centre for Speech Technology Research                  */
/*                     University of Edinburgh, UK                       */
/*                      Copyright (c) 1996,1997                          */
/*                        All Rights Reserved.                           */
/*                                                                       */
/*  Permission to use, copy, modify and distribute this software and its */
/*  documentation for research, educational and individual use only, is  */
/*  hereby granted without fee, subject to the following conditions:     */
/*   1. The code must retain the above copyright notice, this list of    */
/*      conditions and the following disclaimer.                         */
/*   2. Any modifications must be clearly marked as such.                */
/*   3. Original authors' names are not deleted.                         */
/*  This software may not be used for commercial purposes without        */
/*  specific prior written permission from the authors.                  */
/*                                                                       */
/*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        */
/*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      */
/*  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE     */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES    */
/*  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN   */
/*  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,          */
/*  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF       */
/*  THIS SOFTWARE.                                                       */
/*                                                                       */
/*************************************************************************/
/*                     Author :  Alan W Black                            */
/*                     Date   :  May 1996                                */
/*-----------------------------------------------------------------------*/
/*                                                                       */
/* Public declarations for Wagon (CART builder)                          */
/*                                                                       */
/*=======================================================================*/
#ifndef __WAGON_H__
#define __WAGON_H__

#include "EST_simplestats.h"  /* For EST_SuffStats class */
#define wagon_error(WMESS) (cerr << WMESS << endl,exit(-1))

// I get floating point exceptiosn of Alphas when I do any comaprisons
// with HUGE_VAL or FLT_MAX so I'll make my own
#define WGN_HUGE_VAL 1.0e20

class Discretes {
  private:
    int max;
    int next_free;
    EST_Discrete **discretes;
  public:
    Discretes() {max=50;next_free=0;discretes=new EST_Discrete*[max];}
    ~Discretes();
    const int def(const EST_StrList &members);
    EST_Discrete &discrete(const int t) const {return *discretes[t-10];}
    EST_Discrete &operator [] (const int t) const {return *discretes[t-10];}
};

class WIntFlt {
  // Importantly this takes up four bytes only
  private:
    struct intflt { union {int ival; float fval;} u; } val;
  public:
    int &Int(void) { return val.u.ival; }
    float &Flt(void) { return val.u.fval; }
    const int &Int_c(void) const { return val.u.ival; }
    const float &Flt_c(void) const { return val.u.fval; }
};
	
class WVector {
  private:
    int p_size;
    struct WIntFlt *data;
  public:
    WVector() {p_size = 0;}
    WVector(const int &s) {data = new WIntFlt[s]; p_size = s;}
    ~WVector() { if (p_size > 0) delete [] data; }
    int size(void) { return p_size; }
    int get_int_val(int n) const
      { if ((n < 0) || (n >= p_size)) 
        {
          cerr << "WVector index out of range " << n << " " << p_size << "\n";
	  wagon_error("");
	  return data[0].Int();
        }
	else return data[n].Int();
      }
    float get_flt_val(int n) const
      { if ((n < 0) || (n >= p_size))
        {
          cerr << "WVector index out of range " << n << " " << p_size << "\n";
	  wagon_error("");
	  return data[0].Flt(); // wont get here
        }
        else
         return data[n].Flt();
      }
    void set_int_val(int n,int val) const
      { if ((n < 0) || (n >= p_size))
	{
	    cerr << "WVector index out of range " << n << " " 
		<< p_size << "\n";
	    wagon_error("");
	}
	else
	    data[n].Int() = val;
      }
    void set_flt_val(int n,float val) const
      { if ((n < 0) || (n >= p_size))
	{
	    cerr << "WVector index out of range " << n << " " 
		<< p_size << "\n";
	    wagon_error("");
	}
	else
	    data[n].Flt() = val;
      }
    const WIntFlt & operator[] (int n) const
      { if ((n  < 0) || (n >= p_size))
        {
          cerr << "WVector index out of range " << n << " " << p_size << "\n";
          wagon_error("");
  	  return data[0];  // wont get here, but SunCC can't work that out
        }
        else
          return data[n];
      }
	
    WVector & operator =(const WVector &a);
};

typedef EST_TList <WVector *> WVectorList;
typedef EST_TVector <WVector *> WVectorVector;

/* Different types of feature */
enum wn_dtype {wndt_binary, wndt_float, wndt_cluster, wndt_class, wndt_ignore};

class WDataSet {
  private:
    int dlength;
    int types;       // number of types in data;
    WVectorList data;// all vectors
    int *type;       // types for the fields in each vector
    EST_String *name;    // names for the fields
  public:
    WDataSet() {dlength=0; types = wndt_class; data.clear();}
    ~WDataSet() { data.clear();}
    void clear(void) { data.clear();}
    void init(const EST_String& descfname);
    int dtype(const int &i) const {return type[i];}
    WVectorList &get_data(void) {return data;}
    void append (WVector * const &v) { data.append(v); }
    int *get_dtype(void) {return type;}
    EST_String &feat_name(const int &i) const {return name[i];}
    int samples(void) const {return data.length();}
    int width(void) const {return dlength;}
};    
enum wn_oper {wnop_equal, wnop_binary, wnop_greaterthan, 
		  wnop_lessthan, wnop_is, wnop_in, wnop_matches};

class WQuestion {
  private:
    int feature_pos;
    wn_oper op;
    int yes;
    int no;
    EST_Val operand1;
    EST_IList operandl;
    float score;
  public:
    WQuestion() {;}
    WQuestion(const WQuestion &s) 
       { feature_pos=s.feature_pos;
         op=s.op; yes=s.yes; no=s.no; operand1=s.operand1;
	 operandl = s.operandl; score=s.score;}
    ~WQuestion() {;}
    WQuestion(int fp, wn_oper o,EST_Val a)
       { feature_pos=fp; op=o; operand1=a; }
    void set_fp(const int &fp) {feature_pos=fp;}
    void set_oper(const wn_oper &o) {op=o;}
    void set_operand1(const EST_Val &a) {operand1 = a;}
    void set_yes(const int &y) {yes=y;}
    void set_no(const int &n) {no=n;}
    int get_yes(void) const {return yes;}
    int get_no(void) const {return no;}
    const int get_fp(void) const {return feature_pos;}
    const wn_oper get_op(void) const {return op;}
    const EST_Val get_operand1(void) const {return operand1;}
    const EST_IList &get_operandl(void) const {return operandl;}
    const int num_yes(void) const {return yes;}
    const float get_score(void) const {return score;}
    void set_score(const float &f) {score=f;}
    const int ask(const WVector &w) const;
    friend ostream& operator<<(ostream& s, const WQuestion &q);
};

enum wnim_type {wnim_unset, wnim_float, wnim_class, wnim_cluster};

//  Impurity measure for cumulating impurities from set of data
class WImpurity {
  private:
    wnim_type t;
    EST_SuffStats a;
    EST_DiscreteProbDistribution p;
    EST_IList members;            // Maybe there should be a cluster class
    float cluster_impurity();
    float cluster_member_mean(int i);
  public:
    WImpurity() { t=wnim_unset; a.reset(); }
    WImpurity(const WVectorVector &ds);
    float measure(void);
    int samples(void);
    wnim_type type(void) const { return t;}
    void cumulate(const WIntFlt &pv);
    EST_Val value(void);
    EST_DiscreteProbDistribution &pd() { return p; }
    float cluster_distance(int i); // distance i from centre in sds
    int in_cluster(int i);       // distance i from centre < most remote member
    float cluster_ranking(int i);  // position in closeness to centre
    friend ostream& operator<<(ostream &s, WImpurity &imp);
};

class WDlist {
  private:
    float p_score;
    WQuestion p_question;
    EST_String p_token;
    int p_freq;
    int p_samples;
    WDlist *next;
  public:
    WDlist() { next=0; }
    ~WDlist() { if (next != 0) delete next; }
    void set_score(float s) { p_score = s; }
    void set_question(const WQuestion &q) { p_question = q; }
    void set_best(const EST_String &t,int freq, int samples)
	{ p_token = t; p_freq = freq; p_samples = samples;}
    float score() const {return p_score;}
    const EST_String &token(void) const {return p_token;}
    const WQuestion &question() const {return p_question;}
    EST_Val predict(const WVector &w);
    friend WDlist *add_to_dlist(WDlist *l,WDlist *a);
    friend ostream &operator<<(ostream &s, WDlist &d);
};

class WNode {
  private:
    WVectorVector data;
    WQuestion question;
    WImpurity impurity;
    WNode *left;
    WNode *right;
    void print_out(ostream &s, int margin);
    int leaf(void) const { return ((left == 0) || (right == 0)); }
    int pure(void);
  public:
    WNode() { left = right = 0; }
    ~WNode() { if (left != 0) {delete left; left=0;}
	       if (right != 0) {delete right; right=0;} }
    WVectorVector &get_data(void) { return data; }
    void set_subnodes(WNode *l,WNode *r) { left=l; right=r; }
    void set_impurity(const WImpurity &imp) {impurity=imp;}
    void set_question(const WQuestion &q) {question=q;}
    void prune(void);
    void held_out_prune(void);
    WImpurity &get_impurity(void) {return impurity;}
    WQuestion &get_question(void) {return question;}
    EST_Val predict(const WVector &w);
    WNode *predict_node(const WVector &d);
    int samples(void) const { return data.size(); }
    friend ostream& operator<<(ostream &s, WNode &n);
};

void wgn_build_dlist();

extern WDataSet wgn_dataset;
extern ostream *wgn_coutput;
extern Discretes wgn_discretes;
extern EST_String wgn_test_file;
extern EST_String wgn_desc_file;
extern int wgn_min_cluster_size;
extern EST_FMatrix wgn_DistMatrix;

void wgn_load_datadescription(EST_String fname);
void wgn_load_dataset(EST_String fname);
void wgn_build_tree(void);
float wgn_score_question(WQuestion &q, WVectorVector &ds);
void wgn_find_split(WQuestion &q,WVectorVector &ds,
		WVectorVector &y,WVectorVector &n);

extern int wgn_held_out;
extern int wgn_prune;
extern int wgn_quiet;
extern int wgn_csubset;
extern int wgn_vfold;
extern float wgn_float_range_split;
extern float wgn_balance;

#endif /* __WAGON_H__ */
